# coding=utf-8
from .datasets import (TwoMoons as TwoMoons,
                       PimaIndians as PimaIndians,
                       Spiral as Spiral,
                       Gaussians as Gaussians,
                       Iris as Iris)
from .factory import factory as factory